NTSTATUS USB_BulkReadWrite(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PURB	urb;
    ULONG	transferLength;
    NTSTATUS	ntStatus;
    PDEVICE_EXTENSION	deviceExtension;
    PBULK_RW_CONTEXT	rwContext;
    PUSBD_PIPE_INFORMATION	pipeInformation;

    USB_KdPrint (("Hello World I'm USB_BulkReadWrite \n"));

    transferLength = 0;
    deviceExtension = (PDEVICE_EXTENSION) DeviceObject->
							DeviceExtension;
    // iMDL̃p[^jf[^]擾
    if(Irp->MdlAddress) {
        transferLength = MmGetMdlByteCount(Irp->MdlAddress);
    }
    if(!transferLength) {
        USB_KdPrint(("Transfer data length = 0\n"));
        ntStatus = STATUS_SUCCESS;
        goto USB_BulkReadWrite_Exit;
    }

    // oN]pURBmۂCpCv擾
    urb = USB_GetBulkTransferUrb(deviceExtension,Irp,
            transferLength,Irp->MdlAddress,&pipeInformation);
    if(urb == NULL) {
        USB_KdPrint(("Failed to alloc memory for urb\n"));
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        goto USB_BulkReadWrite_Exit;
    }

    // ɌĂяo֐̃p[^f[^mۂ
    rwContext = (PBULK_RW_CONTEXT)ExAllocatePool(NonPagedPool,
                                  sizeof(BULK_RW_CONTEXT));
    if(rwContext == NULL) {
        USB_KdPrint(("Failed to alloc memory for rwContext\n"));
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        goto USB_BulkReadWrite_Exit;
    }

    // [`ɓnp[^ݒ肷
    rwContext->Urb             = urb;
    rwContext->deviceExtension = deviceExtension;
    rwContext->pipeInformation = pipeInformation;

    // USBzXgRg[hCoɏ˗
    IoMarkIrpPending(Irp);
    ntStatus = USB_IoCallDriverToTransferBulkData
							(Irp,rwContext);
    if(!NT_SUCCESS(ntStatus)) {
        USB_KdPrint(("IoCallDriver fails with status %X\n", 
							ntStatus));
        goto USB_BulkReadWrite_Exit;
    }

    // oXfoCXhCȍ҂
    return STATUS_PENDING;

USB_BulkReadWrite_Exit: 	// G[

    if (rwContext) ExFreePool(rwContext);
    if (urb) ExFreePool(urb);

    Irp->IoStatus.Status = ntStatus;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    USB_KdPrint(("USB_BulkReadWrite error (st %x) \n",ntStatus));

    return ntStatus;
}

NTSTATUS
USB_BulkCompletion(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp,
    IN PVOID          Context
    )
{
    // [h/Cg

    ULONG	stageLength;
    NTSTATUS	ntStatus;
    PBULK_RW_CONTEXT	rwContext;
    PDEVICE_EXTENSION	deviceExtension;
    PURB	urb;

    // p[^𕜌
    rwContext = (PBULK_RW_CONTEXT) Context;
    deviceExtension = rwContext->deviceExtension;
    urb = rwContext->Urb;

    ntStatus = Irp->IoStatus.Status;

    USB_KdPrint(("enter USB_BulkCompletion st %x transferLength %d \n",
        ntStatus,urb->UrbBulkOrInterruptTransfer.
							TransferBufferLength));

    // ]߂lɎw肷
    Irp->IoStatus.Information = urb->
				UrbBulkOrInterruptTransfer.TransferBufferLength;

    if (USBD_HALTED(urb->UrbHeader.Status)) {	 // STALLǂ

        // STALLɂȂGh|CgpCv̓Zbg
        // KvɂȂ̂ŁCŊoĂ
        deviceExtension->NeedResetPipe = rwContext->
							pipeInformation;

        Irp->IoStatus.Information = 0;
        USB_KdPrint(("ReadWriteCompletion -	failed with status = %X\n",
            urb->UrbHeader.Status));
    }

    // mۂĂ
    if (rwContext->Urb)
        ExFreePool(rwContext->Urb);
    if (rwContext)
        ExFreePool(rwContext);

    return STATUS_SUCCESS;
}


PURB USB_GetBulkTransferUrb (
    IN PDEVICE_EXTENSION deviceExtension,
    IN PIRP  Irp,
    IN ULONG transferLength,
    IN PMDL  mdl,
    OUT PUSBD_PIPE_INFORMATION *pipeInformation
    )
{
    // oN]URBݒ肵Ė߂lƂ

#define PipeMaximumTransferSize \
    ((PUSBD_PIPE_INFORMATION)*pipeInformation)->MaximumTransferSize

    PIO_STACK_LOCATION		irpStack;
    PURB	urb;
    ULONG	urbFlags;
    BOOLEAN	read;
    ULONG	urbTransferSize = transferLength;

    irpStack = IoGetCurrentIrpStackLocation(Irp);
    read = (irpStack->MajorFunction == IRP_MJ_READ) ?
							 TRUE : FALSE;

    // URBmۂ

    urb = ExAllocatePool(NonPagedPool,
           sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER));
    if(!urb) {
        USB_KdPrint(("Failed to alloc memory for urb\n"));
        return NULL;
    }

    // gpGh|Cg肵ČĂяoɒʒm
    *pipeInformation
        = &deviceExtension->BulkPipeInformation[
            (read) ? \
                USBD_TRANSFER_DIRECTION_IN : 
								USBD_TRANSFER_DIRECTION_OUT ];
    if (transferLength > PipeMaximumTransferSize) {
        //Gh|Cg̍ő]傫
        //Gh|Cg̍ő]ɍ킹
        urbTransferSize = PipeMaximumTransferSize;
    }

    // USBoN]̃tOݒ肷
    urbFlags = USBD_SHORT_TRANSFER_OK;
    if(read) {
       urbFlags |= USBD_TRANSFER_DIRECTION_IN;
       USB_KdPrint(("Read operation\n"));
    }
    else {
       urbFlags |= USBD_TRANSFER_DIRECTION_OUT;
       USB_KdPrint(("Write operation\n"));
    }

    // oN]URB̍쐬
    UsbBuildInterruptOrBulkTransferRequest(
                    urb,
                    sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER),
                    ((PUSBD_PIPE_INFORMATION)*pipeInformation)->PipeHandle,
                    NULL,
                    mdl,
                    urbTransferSize,
                    urbFlags,
                    NULL);

    USB_KdPrint(("USB_GetBulkTransferUrb transferLength %d \n",transferLength));

    return urb;
}

NTSTATUS USB_IoCallDriverToTransferBulkData(
    IN PIRP Irp,
    IN PBULK_RW_CONTEXT rwContext
    )
{
    PIO_STACK_LOCATION  nextStack;

    // ʃfoCXhCoI/OX^bNP[VmۂāC
    // ʃfoCXhCoɈ˗

    USB_KdPrint(("USB_IoCallDriverToTransferBulkData \n"));

    nextStack = IoGetNextIrpStackLocation(Irp);
    nextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
    nextStack->Parameters.Others.Argument1 = rwContext->Urb;
    nextStack->Parameters.DeviceIoControl.IoControlCode =
        IOCTL_INTERNAL_USB_SUBMIT_URB;

    IoSetCompletionRoutine(Irp,
        USB_BulkCompletion,
        rwContext,
        TRUE,
        TRUE,
        TRUE);

    return IoCallDriver(
        rwContext->deviceExtension->StackDeviceObject,Irp);
}
